
/*******************************************************************
 * SLADE - It's a Doom Editor
 * Copyright (C) 2008 Simon Judd
 *
 * Email:       veilofsorrow@gmail.com
 * Web:         http://slade.mancubus.net
 * Filename:    ArchiveManagerPanel.cpp
 * Description: ArchiveManagerPanel class. Basically the UI for the
 *              ArchiveManager class, has a tree of all open
 *              archives and basic file management controls.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *******************************************************************/


/*******************************************************************
 * INCLUDES
 *******************************************************************/
#include "Main.h"
#include "WxStuff.h"
#include "ArchiveManagerPanel.h"
#include "ArchiveManager.h"
#include "ArchivePanel.h"
#include "MapEditorWindow.h"
#include "ZipArchive.h"
#include "BaseResourceArchivesPanel.h"
#include "TextureXEditor.h"
#include "SplashWindow.h"
#include "MainWindow.h"
#include "Icons.h"


/*******************************************************************
 * VARIABLES
 *******************************************************************/
CVAR(Bool, close_archive_with_tab, true, CVAR_SAVE)
CVAR(Int, am_current_tab, 0, CVAR_SAVE)
bool tab_closing = false;	// Hacky workaround to prevent crash on closing a tab when close_archive_with_tab is true


/*******************************************************************
 * EXTERNAL VARIABLES
 *******************************************************************/
EXTERN_CVAR(String, dir_last)


/*******************************************************************
 * WMFILEBROWSER CLASS FUNCTIONS
 *******************************************************************/

/* WMFileBrowser::WMFileBrowser
 * WMFileBrowser class constructor
 *******************************************************************/
WMFileBrowser::WMFileBrowser(wxWindow* parent, ArchiveManagerPanel* wm, int id)
: wxGenericDirCtrl(parent, id, wxDirDialogDefaultFolderStr, wxDefaultPosition, wxDefaultSize, wxDIRCTRL_SHOW_FILTERS,
"Any Supported Archive File (*.wad; *.zip; *.pk3; *.lib; *.dat)|*.wad;*.zip;*.pk3;*.lib;*.dat|Doom Wad files (*.wad)|*.wad|Zip files (*.zip)|*.zip|Pk3 (zip) files (*.pk3)|*.pk3|All Files (*.*)|*.*") {
	// Set the parent
	this->parent = wm;

	// Connect a custom event for when an item in the file tree is activated
	GetTreeCtrl()->Connect(GetTreeCtrl()->GetId(), wxEVT_COMMAND_TREE_ITEM_ACTIVATED, wxTreeEventHandler(WMFileBrowser::onItemActivated));
}

/* WMFileBrowser::~WMFileBrowser
 * WMFileBrowser class destructor
 *******************************************************************/
WMFileBrowser::~WMFileBrowser() {
}

/* WMFileBrowser::onItemActivated
 * Event called when an item in the tree is activated. Opens a file
 * if one is selected.
 *******************************************************************/
void WMFileBrowser::onItemActivated(wxTreeEvent &e) {
	// Get related objects
	wxTreeCtrl* tree = (wxTreeCtrl*) e.GetEventObject();
	WMFileBrowser* browser = (WMFileBrowser*) tree->GetParent();

	// If the selected item has no children (ie it's a file),
	// open it in the archive manager
	if (!tree->ItemHasChildren(e.GetItem()))
		browser->parent->openFile(browser->GetPath());

	e.Skip();
}


/*******************************************************************
 * ARCHIVEMANAGERPANEL CLASS FUNCTIONS
 *******************************************************************/

/* ArchiveManagerPanel::ArchiveManagerPanel
 * ArchiveManagerPanel class constructor
 *******************************************************************/
ArchiveManagerPanel::ArchiveManagerPanel(wxWindow *parent, wxAuiNotebook* nb_archives)
: wxPanel(parent, -1) {
	notebook_archives = nb_archives;

	// Create main sizer
	wxBoxSizer *vbox = new wxBoxSizer(wxVERTICAL);
	SetSizer(vbox);

	// Create/setup tabs
	notebook_tabs = new wxAuiNotebook(this, -1, wxDefaultPosition, wxDefaultSize, wxAUI_NB_TOP|wxAUI_NB_TAB_SPLIT|wxAUI_NB_TAB_MOVE|wxAUI_NB_SCROLL_BUTTONS|wxAUI_NB_WINDOWLIST_BUTTON);
	vbox->Add(notebook_tabs, 1, wxEXPAND | wxALL, 4);

	// Open archives & maps list
	wxPanel *panel_am = new wxPanel(notebook_tabs);
	notebook_tabs->AddPage(panel_am, "Archives", true);

	// Create/setup archive list
	wxBoxSizer *box_am = new wxBoxSizer(wxVERTICAL);
	panel_am->SetSizer(box_am);
	box_am->Add(new wxStaticText(panel_am, -1, "Open Archives:"), 0, wxEXPAND | wxALL, 4);
	list_archives = new ListView(panel_am, -1);
	box_am->Add(list_archives, 1, wxEXPAND | wxLEFT | wxRIGHT | wxBOTTOM, 4);
	refreshArchiveList();

	// Create/setup map list
	box_am->Add(new wxStaticText(panel_am, -1, "Maps:"), 0, wxEXPAND | wxALL, 4);
	list_maps = new wxListCtrl(panel_am, -1, wxDefaultPosition, wxSize(-1, 128), wxLC_LIST | wxLC_SINGLE_SEL);
	box_am->Add(list_maps, 0, wxEXPAND | wxLEFT | wxRIGHT | wxBOTTOM, 4);

	// Create/setup file browser
	file_browser = new WMFileBrowser(notebook_tabs, this, -1);
	notebook_tabs->AddPage(file_browser, _("File Browser"));

	// Create/setup Archive context menu
	menu_context_open = new wxMenu();
	menu_context_open->Append(MENU_SAVE, "Save", "Save the selected Archive(s)");
	menu_context_open->Append(MENU_SAVEAS, "Save As", "Save the selected Archive(s) to a new file(s)");
	menu_context_open->Append(MENU_CLOSE, "Close", "Close the selected Archive(s)");

	// Create/setup recent files list and menu
	menu_recent = new wxMenu();
	wxPanel *panel_rf = new wxPanel(notebook_tabs);
	wxBoxSizer *box_rf = new wxBoxSizer(wxVERTICAL);
	panel_rf->SetSizer(box_rf);
	box_rf->Add(new wxStaticText(panel_rf, -1, "Recent Files:"), 0, wxEXPAND | wxALL, 4);
	list_recent = new ListView(panel_rf, -1);
	box_rf->Add(list_recent, 1, wxEXPAND | wxLEFT | wxRIGHT | wxBOTTOM, 4);
	refreshRecentFileList();
	notebook_tabs->AddPage(panel_rf, "Recent Files", true);

	// Create/setup Archive context menu
	menu_context_recent = new wxMenu();
	menu_context_recent->Append(MENU_OPEN, "Open", "Open the selected Archive(s)");
	menu_context_recent->Append(MENU_REMOVE, "Remove", "Remove the selected Archive(s) from the Recent list");

	// Create/setup bookmarks tab
	wxPanel *panel_bm = new wxPanel(notebook_tabs);
	wxBoxSizer *box_bm = new wxBoxSizer(wxVERTICAL);
	panel_bm->SetSizer(box_bm);
	box_bm->Add(new wxStaticText(panel_bm, -1, "Bookmarks:"), 0, wxEXPAND | wxALL, 4);
	list_bookmarks = new ListView(panel_bm, -1);
	box_bm->Add(list_bookmarks, 1, wxEXPAND | wxLEFT | wxRIGHT | wxBOTTOM, 4);
	refreshBookmarkList();
	notebook_tabs->AddPage(panel_bm, "Bookmarks", true);

	// Set current tab
	notebook_tabs->SetSelection(am_current_tab);

	// Create/setup Archive context menu
	menu_context_bookmarks = new wxMenu();
	menu_context_bookmarks->Append(MENU_GO, "Go To", "Go to the chosen bookmark");
	menu_context_bookmarks->Append(MENU_DELETE, "Remove", "Remove the selected bookmarks from the list");

	// Bind events
	list_archives->Bind(wxEVT_COMMAND_LIST_ITEM_SELECTED, &ArchiveManagerPanel::onListArchivesChanged, this);
	list_archives->Bind(wxEVT_COMMAND_LIST_ITEM_ACTIVATED, &ArchiveManagerPanel::onListArchivesActivated, this);
	list_archives->Bind(wxEVT_COMMAND_LIST_ITEM_RIGHT_CLICK, &ArchiveManagerPanel::onListArchivesRightClick, this);
	list_recent->Bind(wxEVT_COMMAND_LIST_ITEM_ACTIVATED, &ArchiveManagerPanel::onListRecentActivated, this);
	list_recent->Bind(wxEVT_COMMAND_LIST_ITEM_RIGHT_CLICK, &ArchiveManagerPanel::onListRecentRightClick, this);
	list_bookmarks->Bind(wxEVT_COMMAND_LIST_ITEM_ACTIVATED, &ArchiveManagerPanel::onListBookmarksActivated, this);
	list_bookmarks->Bind(wxEVT_COMMAND_LIST_ITEM_RIGHT_CLICK, &ArchiveManagerPanel::onListBookmarksRightClick, this);
	list_maps->Bind(wxEVT_COMMAND_LISTBOX_SELECTED, &ArchiveManagerPanel::onListMapsChanged, this);
	list_maps->Bind(wxEVT_COMMAND_LIST_ITEM_ACTIVATED, &ArchiveManagerPanel::onListMapsActivated, this);
	notebook_archives->Bind(wxEVT_COMMAND_AUINOTEBOOK_PAGE_CHANGED, &ArchiveManagerPanel::onArchiveTabChanged, this);
	notebook_archives->Bind(wxEVT_COMMAND_AUINOTEBOOK_PAGE_CLOSE, &ArchiveManagerPanel::onArchiveTabClose, this);
	notebook_tabs->Bind(wxEVT_COMMAND_AUINOTEBOOK_PAGE_CHANGED, &ArchiveManagerPanel::onAMTabChanged, this);
	Bind(wxEVT_COMMAND_MENU_SELECTED, &ArchiveManagerPanel::onMenu, this, MENU_SAVE, MENU_END);

	// Listen to the ArchiveManager
	listenTo(theArchiveManager);

	// Init layout
	Layout();
}

/* ArchiveManagerPanel::~ArchiveManagerPanel
 * ArchiveManagerPanel class destructor
 *******************************************************************/
ArchiveManagerPanel::~ArchiveManagerPanel() {
	if (menu_context_open) delete menu_context_open;
	if (menu_context_recent) delete menu_context_recent;
	if (menu_context_bookmarks) delete menu_context_bookmarks;
}

/* ArchiveManagerPanel::refreshRecentFileList
 * Clears and rebuilds the recent file list in the menu and the tab
 *******************************************************************/
void ArchiveManagerPanel::refreshRecentFileList() {
	// Clear the list
	list_recent->ClearAll();

	// Clear menu; needs to do with a count down rather than up
	// otherwise the following elements are not properly removed
	for (unsigned a = menu_recent->GetMenuItemCount(); a > 0; a--)
		menu_recent->Destroy(MainWindow::MENU_RECENT_1 + a - 1);

	// Add columns
	list_recent->InsertColumn(0, "Filename");
	list_recent->InsertColumn(1, "Path");

	// Add each recent archive (same logic as the recent files submenu)
	list_recent->enableSizeUpdate(false);
	for (unsigned a = 0; a < theArchiveManager->numRecentFiles(); a++) {
		list_recent->addItem(a, wxEmptyString);
		updateRecentListItem(a);

		if (a < 8)
			menu_recent->Append(MainWindow::MENU_RECENT_1+a, theArchiveManager->recentFile(a));
	}

	// Update size
	list_recent->enableSizeUpdate(true);
	list_recent->updateSize();
}

/* ArchiveManagerPanel::refreshArchiveList
 * Clears and rebuilds the open archives list
 *******************************************************************/
void ArchiveManagerPanel::refreshArchiveList() {
	// Clear the list
	list_archives->ClearAll();

	// Add columns
	list_archives->InsertColumn(0, "Filename");
	list_archives->InsertColumn(1, "Path");

	// Add each archive that is opened in the ArchiveManager
	list_archives->enableSizeUpdate(false);
	for (int a = 0; a < theArchiveManager->numArchives(); a++) {
		list_archives->addItem(a, wxEmptyString);
		updateOpenListItem(a);
	}

	// Update size
	list_archives->enableSizeUpdate(true);
	list_archives->updateSize();
}

/* ArchiveManagerPanel::refreshAllTabs
 * Refreshes all open archive tabs
 *******************************************************************/
void ArchiveManagerPanel::refreshAllTabs() {
	// Go through tabs
	for (unsigned a = 0; a < notebook_archives->GetPageCount(); a++) {
		wxWindow* tab = notebook_archives->GetPage(a);

		// Refresh if it's an archive panel
		if (isArchivePanel(a))
			((ArchivePanel*)tab)->refreshPanel();
	}
}

/* ArchiveManagerPanel::populateMapList
 * Adds all maps in <archive> to the maps list widget
 *******************************************************************/
void ArchiveManagerPanel::populateMapList(Archive* archive) {
	// Clear current maps list
	list_maps->ClearAll();

	// Do nothing if no archive was given
	if (!archive)
		return;

	// Get the list of maps in the selected archive
	vector<Archive::mapdesc_t> maps = archive->detectMaps();

	// Go through the list and add maps
	for (int a = 0; a < (int) maps.size(); a++) {
		// Setup map name string
		string name;

		// Add map format to name string
		switch(maps[a].format) {
			case MAP_DOOM:		name = "(D) "; break;
			case MAP_HEXEN:		name = "(H) "; break;
			case MAP_DOOM64:	name = "(K) "; break;
			case MAP_UDMF:		name = "(U) "; break;
		}

		// Add map name to string
		name += maps[a].name;

		// Add the list item
		wxListItem li;
		li.SetText(name);
		li.SetId(a);
		list_maps->InsertItem(li);
	}
}

/* ArchiveManagerPanel::updateOpenListItem
 * Updates the archive list item at <index>
 *******************************************************************/
void ArchiveManagerPanel::updateOpenListItem(int index) {
	Archive* archive = theArchiveManager->getArchive(index);

	if (!archive)
		return;

	// Get path as wxFileName for processing
	wxFileName fn(archive->getFilename());

	// Set item name
	list_archives->setItemText(index, 0, fn.GetFullName());
	list_archives->setItemText(index, 1, fn.GetPath());

	// Set item status colour
	if (archive->canSave()) {
		if (archive->isModified())
			list_archives->setItemStatus(index, LV_STATUS_MODIFIED);
		else
			list_archives->setItemStatus(index, LV_STATUS_NORMAL);
	}
	else
		list_archives->setItemStatus(index, LV_STATUS_NEW);
}

/* ArchiveManagerPanel::updateRecentListItem
 * Updates the recent file list item at <index>
 *******************************************************************/
void ArchiveManagerPanel::updateRecentListItem(int index) {

	// Get path as wxFileName for processing
	wxFileName fn(theArchiveManager->recentFile(index));

	// Set item name
	list_recent->setItemText(index, 0, fn.GetFullName());
	list_recent->setItemText(index, 1, fn.GetPath());
}

/* ArchiveManagerPanel::isArchivePanel
 * Checks if the tab at [tab_index] is an ArchivePanel. Returns true
 * if it is, false if not
 *******************************************************************/
bool ArchiveManagerPanel::isArchivePanel(int tab_index) {
	// Check that tab index is in range
	if ((unsigned)tab_index >= notebook_archives->GetPageCount())
		return false;

	// Check the page's name
	if (!notebook_archives->GetPage(tab_index)->GetName().CmpNoCase("archive"))
		return true;
	else
		return false;
}

/* ArchiveManagerPanel::getArchive
 * Returns the archive associated with the archive tab at [tab_index]
 * or NULL if the index is invalid or the tab isn't an archive panel
 *******************************************************************/
Archive* ArchiveManagerPanel::getArchive(int tab_index) {
	// Check the index is valid
	if (tab_index < 0 || (unsigned)tab_index >= notebook_archives->GetPageCount())
		return NULL;

	// Check the specified tab is actually an archive tab
	if (!isArchivePanel(tab_index))
		return NULL;

	// Get the archive associated with the tab
	ArchivePanel* ap = (ArchivePanel*)notebook_archives->GetPage(tab_index);
	return ap->getArchive();
}

/* ArchiveManagerPanel::currentTabIndex
 * Returns the index of the currently open tab
 *******************************************************************/
int ArchiveManagerPanel::currentTabIndex() {
	return notebook_archives->GetSelection();
}

/* ArchiveManagerPanel::currentArchive
 * Returns the currently 'open' archive - the archive associated
 * with the current ArchivePanel tab. Returns NULL if the current tab
 * isn't an ArchivePanel
 *******************************************************************/
Archive* ArchiveManagerPanel::currentArchive() {
	// Get current tab index
	int selected = notebook_archives->GetSelection();

	// Check it's an archive tab
	if (!isArchivePanel(selected))
		return NULL;

	// Get the archive associated with the panel
	ArchivePanel* ap = (ArchivePanel*)notebook_archives->GetPage(selected);
	return ap->getArchive();
}

/* ArchiveManagerPanel::currentPanel
 * Returns the current panel tab
 *******************************************************************/
wxWindow* ArchiveManagerPanel::currentPanel() {
	return notebook_archives->GetPage(notebook_archives->GetSelection());
}

/* ArchiveManagerPanel::currentArea
 * Returns the current panel tab
 *******************************************************************/
EntryPanel* ArchiveManagerPanel::currentArea() {
	// Get current tab index
	int selected = notebook_archives->GetSelection();

	// Check it's an archive tab
	if (!isArchivePanel(selected))
		return NULL;

	ArchivePanel* ap = (ArchivePanel*)notebook_archives->GetPage(selected);
	return ap->currentArea();
}

/* ArchiveManagerPanel::currentEntry
 * Returns the currently open entry in the current archive panel
 * (if any)
 *******************************************************************/
ArchiveEntry* ArchiveManagerPanel::currentEntry() {
	// Get current tab index
	int selected = notebook_archives->GetSelection();

	// Check it's an archive tab
	if (!isArchivePanel(selected))
		return NULL;

	// Get the archive panel
	ArchivePanel* ap = (ArchivePanel*)notebook_archives->GetPage(selected);
	return ap->currentEntry();
}

/* ArchiveManagerPanel::currentEntrySelection
 * Returns a list of all selected entries in the current archive
 * panel
 *******************************************************************/
vector<ArchiveEntry*> ArchiveManagerPanel::currentEntrySelection() {
	// Get current tab index
	int selected = notebook_archives->GetSelection();

	// Check it's an archive tab
	if (!isArchivePanel(selected))
		return vector<ArchiveEntry*>();

	// Get the archive panel
	ArchivePanel* ap = (ArchivePanel*)notebook_archives->GetPage(selected);
	return ap->currentEntries();
}

/* ArchiveManagerPanel::openTab
 * Opens a new tab for the archive at <archive_index> in the archive
 * manager
 *******************************************************************/
void ArchiveManagerPanel::openTab(int archive_index) {
	Archive* archive = theArchiveManager->getArchive(archive_index);
	if (archive) {
		openTab(archive);
	}
}

/* ArchiveManagerPanel::openTab
 * Opens a new tab for the archive
 *******************************************************************/
void ArchiveManagerPanel::openTab(Archive* archive) {
	if (archive) {
		// Go through all tabs
		for (size_t a = 0; a < notebook_archives->GetPageCount(); a++) {
			// Check page type is "archive"
			if (notebook_archives->GetPage(a)->GetName().CmpNoCase("archive"))
				continue;

			// Check for archive match
			ArchivePanel* ap = (ArchivePanel*)notebook_archives->GetPage(a);
			if (ap->getArchive() == archive) {
				// Selected archive is already open in a tab, so switch to this tab
				notebook_archives->SetSelection(a);
				ap->focusEntryList();
				return;
			}
		}

		// If tab isn't already open, open a new one
		ArchivePanel* wp = new ArchivePanel(notebook_archives, archive);

		// Determine icon
		string icon = "e_archive";
		if (archive->getType() == ARCHIVE_WAD)
			icon = "e_wad";
		else if (archive->getType() == ARCHIVE_ZIP)
			icon = "e_zip";

		notebook_archives->AddPage(wp, archive->getFilename(false), false);
		notebook_archives->SetSelection(notebook_archives->GetPageCount() - 1);
		notebook_archives->SetPageBitmap(notebook_archives->GetPageCount() - 1, getIcon(icon));
		wp->SetName("archive");
		wp->Show(true);
		wp->SetFocus();
		wp->focusEntryList();
	}
}

/* ArchiveManagerPanel::closeTab
 * Closes the archive editor tab for the archive at <archive_index>
 * in the archive manager
 *******************************************************************/
void ArchiveManagerPanel::closeTab(int archive_index) {
	// Don't close if a tab is already closing
	if (tab_closing)
		return;

	Archive* archive = theArchiveManager->getArchive(archive_index);

	if (archive) {
		// Go through all tabs
		for (size_t a = 0; a < notebook_archives->GetPageCount(); a++) {
			// Check page type is "archive"
			if (notebook_archives->GetPage(a)->GetName().CmpNoCase("archive"))
				continue;

			// Check for archive match
			ArchivePanel* ap = (ArchivePanel*)notebook_archives->GetPage(a);
			if (ap->getArchive() == archive) {
				// Close the tab
				notebook_archives->DeletePage(a);

				return;
			}
		}
	}
}

/* ArchiveManagerPanel::openTextureTab
 * Opens a new texture editor tab for the archive at <archive_index>
 * in the archive manager
 *******************************************************************/
void ArchiveManagerPanel::openTextureTab(int archive_index) {
	Archive* archive = theArchiveManager->getArchive(archive_index);

	if (archive) {
		// Go through all tabs
		for (size_t a = 0; a < notebook_archives->GetPageCount(); a++) {
			// Check page type is "texture"
			if (notebook_archives->GetPage(a)->GetName().CmpNoCase("texture"))
				continue;

			// Check for archive match
			TextureXEditor* txed = (TextureXEditor*)notebook_archives->GetPage(a);
			if (txed->getArchive() == archive) {
				// Selected archive already has its texture editor open, so show that tab
				notebook_archives->SetSelection(a);
				return;
			}
		}

		// If tab isn't already open, open a new one
		TextureXEditor* txed = new TextureXEditor(notebook_archives);
		txed->Show(false);
		if (!txed->openArchive(archive)) {
			delete txed;
			return;
		}

		notebook_archives->AddPage(txed, s_fmt("Texture Editor (%s)", archive->getFilename(false).c_str()), true);
		notebook_archives->SetPageBitmap(notebook_archives->GetPageCount() - 1, getIcon("e_texturex"));
		txed->SetName("texture");
		txed->Show(true);
		// Select the new tab
		for (size_t a = 0; a < notebook_archives->GetPageCount(); a++) {
			if (notebook_archives->GetPage(a) == txed) {
				notebook_archives->SetSelection(a);
				return;
			}
		}
	}
}

/* ArchiveManagerPanel::closeTextureTab
 * Closes the texture editor tab for the archive at <archive_index>
 * in the archive manager
 *******************************************************************/
void ArchiveManagerPanel::closeTextureTab(int archive_index) {
	Archive* archive = theArchiveManager->getArchive(archive_index);

	if (archive) {
		// Go through all tabs
		for (size_t a = 0; a < notebook_archives->GetPageCount(); a++) {
			// Check page type is "texture"
			if (notebook_archives->GetPage(a)->GetName().CmpNoCase("texture"))
				continue;

			// Check for archive match
			TextureXEditor* txed = (TextureXEditor*)notebook_archives->GetPage(a);
			if (txed->getArchive() == archive) {
				// Close the tab
				notebook_archives->DeletePage(a);

				return;
			}
		}
	}
}

/* ArchiveManagerPanel::openFile
 * Opens an archive and initialises the UI for it
 *******************************************************************/
void ArchiveManagerPanel::openFile(string filename) {
	// Show splash screen
	theSplashWindow->show("Opening Archive...", true);

	// test
	wxStopWatch sw;
	sw.Start();

	// Open the file in the archive manager
	Archive* new_archive = theArchiveManager->openArchive(filename);

	sw.Pause();
	wxLogMessage("Opening took %d ms", (int)sw.Time());

	// Hide splash screen
	theSplashWindow->hide();

	// Check that the archive opened ok
	if (!new_archive) {
		// If archive didn't open ok, show error message
		wxMessageBox(s_fmt("Error opening %s:\n%s", filename.c_str(), Global::error.c_str()), "Error", wxICON_ERROR);
	}
}

/* ArchiveManagerPanel::openFiles
 * Opens each file in a supplied array of filenames
 *******************************************************************/
void ArchiveManagerPanel::openFiles(wxArrayString& files) {
	// Go through each filename in the array
	for (int a = 0; a < (int) files.size(); a++) {
		// Open the archive
		openFile(files[a]);
	}
}

/* ArchiveManagerPanel::closeAll
 * Closes all currently open archives
 *******************************************************************/
bool ArchiveManagerPanel::closeAll() {
	for (int a = 0; a < theArchiveManager->numArchives(); a++) {
		if (!closeArchive(theArchiveManager->getArchive(a)))
			return false;
	}

	return true;
}

/* ArchiveManagerPanel::saveAll
 * Saves all currently open archives
 *******************************************************************/
void ArchiveManagerPanel::saveAll() {
	// Go through all open archives
	for (int a = 0; a < theArchiveManager->numArchives(); a++) {
		// Get the archive to be saved
		Archive* archive = theArchiveManager->getArchive(a);

		if (archive->canSave()) {
			// Save the archive if possible
			if (!archive->save()) {
				// If there was an error pop up a message box
				wxMessageBox(s_fmt("Error: %s", Global::error.c_str()), "Error", wxICON_ERROR);
			}
		}
		else {
			// If the archive is newly created, do Save As instead

			// Popup file save dialog
			string formats = archive->getFileExtensionString();
			string filename = wxFileSelector("Save Archive " + archive->getFilename(false) + " As", dir_last, "", wxEmptyString, formats, wxFD_SAVE | wxFD_OVERWRITE_PROMPT);

			// Check a filename was selected
			if (!filename.empty()) {
				// Save the archive
				if (!archive->save(filename)) {
					// If there was an error pop up a message box
					wxMessageBox(s_fmt("Error: %s", Global::error.c_str()), "Error", wxICON_ERROR);
				}

				// Save 'dir_last'
				wxFileName fn(filename);
				dir_last = fn.GetPath(true);
			}
		}
	}
}

/* ArchiveManagerPanel::createNewArchive
 * Creates a new archive of the given type and opens it in a tab
 *******************************************************************/
void ArchiveManagerPanel::createNewArchive(uint8_t type) {
	Archive* new_archive = theArchiveManager->newArchive(type);

	if (new_archive) {
		openTab(theArchiveManager->archiveIndex(new_archive));
	}
}

/* ArchiveManagerPanel::saveEntryChanges
 * If there are any unsaved entry changes in [archive]'s ArchivePanel
 * tab, saves the changes (or not, depending on user settings)
 *******************************************************************/
bool ArchiveManagerPanel::saveEntryChanges(Archive* archive) {
	// Go through tabs
	for (size_t a = 0; a < notebook_archives->GetPageCount(); a++) {
		// Check page type is "archive"
		if (notebook_archives->GetPage(a)->GetName().CmpNoCase("archive"))
			continue;

		// Check for archive match
		ArchivePanel* ap = (ArchivePanel*)notebook_archives->GetPage(a);
		if (ap->getArchive() == archive) {
			// Save entry changes
			return ap->saveEntryChanges();
		}
	}

	return false;
}

/* ArchiveManagerPanel::saveArchive
 * Saves [archive] to disk, opens a file dialog if necessary
 *******************************************************************/
bool ArchiveManagerPanel::saveArchive(Archive* archive) {
	// Check for unsaved entry changes
	saveEntryChanges(archive);

	if (archive->isOnDisk()) {
		// Save the archive if possible
		if (!archive->save()) {
			// If there was an error pop up a message box
			wxMessageBox(s_fmt("Error: %s", Global::error.c_str()), "Error", wxICON_ERROR);
			return false;
		}

		return true;
	}
	else
		return saveArchiveAs(archive);	// If the archive is newly created, do Save As instead
}

/* ArchiveManagerPanel::saveArchive
 * Saves [archive] to disk under a different filename, opens
 * a file dialog to select the new name/path
 *******************************************************************/
bool ArchiveManagerPanel::saveArchiveAs(Archive* archive) {
	// Check for unsaved entry changes
	saveEntryChanges(archive);

	// Popup file save dialog
	string formats = archive->getFileExtensionString();
	string filename = wxFileSelector("Save Archive " + archive->getFilename(false) + " As", dir_last, "", wxEmptyString, formats, wxFD_SAVE | wxFD_OVERWRITE_PROMPT);

	// Check a filename was selected
	if (!filename.empty()) {
		// Save the archive
		if (!archive->save(filename)) {
			// If there was an error pop up a message box
			wxMessageBox(s_fmt("Error: %s", Global::error.c_str()), "Error", wxICON_ERROR);
			return false;
		}

		// Save 'dir_last'
		wxFileName fn(filename);
		dir_last = fn.GetPath(true);
	}

	return true;
}

/* ArchiveManagerPanel::closeArchive
 * Checks for any unsaved changes and prompts the user to save if
 * needed before closing [archive]
 *******************************************************************/
bool ArchiveManagerPanel::closeArchive(Archive* archive) {
	// Check for NULL pointers -- this can happen, for example,
	// with onArchiveTabClose() when closing a texture editor tab.
	if (!archive)
		return false;

	// Check for unsaved entry changes
	saveEntryChanges(archive);

	// If the archive has unsaved changes, prompt to save
	if (archive->isModified()) {
		wxMessageDialog md(this, s_fmt("Save changes to %s?", archive->getFilename(false)), "Unsaved Changes", wxYES_NO|wxCANCEL);
		int result = md.ShowModal();
		if (result == wxID_YES) {
			// User selected to save
			if (!saveArchive(archive))
				return false;	// Unsuccessful save, don't close
		}
		else if (result == wxID_CANCEL)
			return false;	// User selected cancel, don't close the archive
	}

	// Close the archive
	return theArchiveManager->closeArchive(archive);
}

/* ArchiveManagerPanel::getSelectedArchives
 * Gets a list of indices of all selected archive list items
 *******************************************************************/
vector<int> ArchiveManagerPanel::getSelectedArchives() {
	vector<int> ret;

	// Go through all wad list items
	long item = -1;
	while (true) {
		// Get the next item in the list that is selected
		item = list_archives->GetNextItem(item, wxLIST_NEXT_ALL, wxLIST_STATE_SELECTED);

		// If -1 then none were selected
		if (item == -1)
			break;

		// Otherwise add the selected item to the vector
		ret.push_back(item);
	}

	return ret;
}

/* ArchiveManagerPanel::getSelectedFiles
 * Gets a list of indices of all selected recent file list items
 *******************************************************************/
vector<int> ArchiveManagerPanel::getSelectedFiles() {
	vector<int> ret;

	// Go through all wad list items
	long item = -1;
	while (true) {
		// Get the next item in the list that is selected
		item = list_recent->GetNextItem(item, wxLIST_NEXT_ALL, wxLIST_STATE_SELECTED);

		// If -1 then none were selected
		if (item == -1)
			break;

		// Otherwise add the selected item to the vector
		ret.push_back(item);
	}

	return ret;
}

/* ArchiveManagerPanel::getSelectedBookmarks
 * Gets a list of indices of all selected bookmark list items
 *******************************************************************/
vector<int> ArchiveManagerPanel::getSelectedBookmarks() {
	vector<int> ret;

	// Go through all wad list items
	long item = -1;
	while (true) {
		// Get the next item in the list that is selected
		item = list_bookmarks->GetNextItem(item, wxLIST_NEXT_ALL, wxLIST_STATE_SELECTED);

		// If -1 then none were selected
		if (item == -1)
			break;

		// Otherwise add the selected item to the vector
		ret.push_back(item);
	}

	return ret;
}

/* ArchiveManagerPanel::onAnnouncement
 * Called when an announcement is recieved from the Archive Manager
 *******************************************************************/
void ArchiveManagerPanel::onAnnouncement(Announcer* announcer, string event_name, MemChunk& event_data) {
	// Reset event data for reading
	event_data.seek(0, SEEK_SET);

	// If an archive is about to be closed
	if (event_name == "archive_closing") {
		int32_t index = -1;
		event_data.read(&index, 4);

		// Close any related tabs
		closeTextureTab(index);
		closeTab(index);
	}

	// If an archive was closed
	if (event_name == "archive_closed") {
		int32_t index = -1;
		event_data.read(&index, 4);
		refreshArchiveList();
		populateMapList(NULL);
	}

	// If an archive was added
	if (event_name == "archive_added") {
		int index = theArchiveManager->numArchives() - 1;
		list_archives->addItem(index, wxEmptyString);
		updateOpenListItem(index);
	}

	// If an archive was opened
	if (event_name == "archive_opened") {
		uint32_t index = -1;
		event_data.read(&index, 4);
		openTab(index);
	}

	// If an archive was saved
	if (event_name == "archive_saved") {
		int32_t index = -1;
		event_data.read(&index, 4);
		updateOpenListItem(index);
	}

	// If an archive was modified
	if (event_name == "archive_modified") {
		int32_t index = -1;
		event_data.read(&index, 4);
		updateOpenListItem(index);
	}

	// If a texture editor is to be opened
	if (event_name == "open_tex_editor") {
		uint32_t index = 0;
		event_data.read(&index, 4);
		openTextureTab(index);
	}

	// If the recent files list has changed
	if (event_name == "recent_files_changed") {
		refreshRecentFileList();
	}

	// If the bookmarks list has changed
	if (event_name == "bookmarks_changed") {
		refreshBookmarkList();
	}
}

/* ArchiveManagerPanel::saveSelection
 * Saves the currently selected archive(s) in the list
 *******************************************************************/
void ArchiveManagerPanel::saveSelection() {
	// Get the list of selected archives
	vector<int> selection = getSelectedArchives();

	// Don't continue if there are no selected items
	if (selection.size() == 0)
		return;

	// Go through the selection
	for (size_t a = 0; a < selection.size(); a++) {
		// Get the archive to be saved
		Archive* archive = theArchiveManager->getArchive(selection[a]);

		saveArchive(archive);
	}
}

/* ArchiveManagerPanel::saveSelectionAs
 * Saves the currently selected archive(s) in the list as new file(s)
 *******************************************************************/
void ArchiveManagerPanel::saveSelectionAs() {
	// Get the list of selected archives
	vector<int> selection = getSelectedArchives();

	// Don't continue if there are no selected items
	if (selection.size() == 0)
		return;

	// Go through the selection
	for (size_t a = 0; a < selection.size(); a++) {
		// Get the archive
		Archive* archive = theArchiveManager->getArchive(selection[a]);
		saveArchiveAs(archive);
	}

	refreshArchiveList();
}

/* ArchiveManagerPanel::closeSelection
 * Closes the currently selected archive(s) in the list
 *******************************************************************/
bool ArchiveManagerPanel::closeSelection() {
	// Get the list of selected list items
	vector<int> selection = getSelectedArchives();

	// Don't continue if there are no selected items
	if (selection.size() == 0)
		true;

	// Get the list of selected archives
	vector<Archive*> selected_archives;
	for (size_t a = 0; a < selection.size(); a++)
		selected_archives.push_back(theArchiveManager->getArchive(selection[a]));

	// Close all selected archives
	bool all_closed = true;
	for (size_t a = 0; a < selected_archives.size(); a++) {
		if (!closeArchive(selected_archives[a]))
			all_closed = false;
	}

	return all_closed;
}

/* ArchiveManagerPanel::openSelection
 * Open the currently selected archive(s) in the recent file list
 *******************************************************************/
void ArchiveManagerPanel::openSelection() {
	// Get the list of selected list items
	vector<int> selection = getSelectedFiles();

	// Don't continue if there are no selected items
	if (selection.size() == 0)
		return;

	// Get the list of selected archives
	vector<string> selected_archives;
	for (size_t a = 0; a < selection.size(); a++)
		selected_archives.push_back(theArchiveManager->recentFile(selection[a]));

	// Open all selected archives
	for (size_t a = 0; a < selected_archives.size(); a++)
		theArchiveManager->openArchive(selected_archives[a]);
}

/* ArchiveManagerPanel::openSelection
 * Remove the currently selected archive(s) from the recent file list
 *******************************************************************/
void ArchiveManagerPanel::removeSelection() {
	// Get the list of selected list items
	vector<int> selection = getSelectedFiles();

	// Don't continue if there are no selected items
	if (selection.size() == 0)
		return;

	// Prepare list of kept archives
	int numfiles = theArchiveManager->numRecentFiles();
	bool * okay = new bool[numfiles];
	for (int a = 0; a < (signed)numfiles; ++a)
		okay[a] = true;

	// Deselect the files that are being removed
	for (int a = 0; a < (signed)selection.size(); ++a) {
		if (selection[a] < numfiles)
			okay[selection[a]] = false;
	}

	// Get the list of kept files
	vector<string> kept_files;
	for (int a = 0; a < numfiles; a++)
		if (okay[a])
			kept_files.push_back(theArchiveManager->recentFile(a));

	delete[] okay;

	// Renew the list of recent files
	theArchiveManager->addRecentFiles(kept_files);
}

void ArchiveManagerPanel::handleAction(int menu_id) {
	// *************************************************************
	// FILE MENU
	// *************************************************************

	// File->New Wad
	if (menu_id == MainWindow::MENU_FILE_NEWWAD)
		createNewArchive(ARCHIVE_WAD);

	// File->New Zip
	else if (menu_id == MainWindow::MENU_FILE_NEWZIP)
		createNewArchive(ARCHIVE_ZIP);

	// File->Open
	else if (menu_id == MainWindow::MENU_FILE_OPEN) {
		// Create extensions string
		string extensions = theArchiveManager->getArchiveExtensionsString();

		// Open a file browser dialog that allows multiple selection
		// and filters by wad, zip and pk3 file extensions
		wxFileDialog dialog_open(this, "Choose file(s) to open", dir_last, wxEmptyString, extensions, wxFD_OPEN|wxFD_MULTIPLE|wxFD_FILE_MUST_EXIST, wxDefaultPosition);

		// Run the dialog & check that the user didn't cancel
		if (dialog_open.ShowModal() == wxID_OK) {
			wxBeginBusyCursor();

			// Get an array of selected filenames
			wxArrayString files;
			dialog_open.GetPaths(files);

			// Open them
			openFiles(files);

			wxEndBusyCursor();

			// Save 'dir_last'
			dir_last = dialog_open.GetDirectory();
		}
	}

	// File->Recent
	else if (menu_id >= MainWindow::MENU_RECENT_1 && menu_id <= MainWindow::MENU_RECENT_8) {
		// Get recent file index
		unsigned index = menu_id - MainWindow::MENU_RECENT_1;

		// Open it
		openFile(theArchiveManager->recentFile(index));
	}

	// File->Save All
	else if (menu_id == MainWindow::MENU_FILE_SAVEALL)
		saveAll();

	// File->Close All
	else if (menu_id == MainWindow::MENU_FILE_CLOSEALL)
		closeAll();

	// File->Close
	else if (menu_id == MainWindow::MENU_FILE_CLOSE)
		closeArchive(currentArchive());

	else {
		// Check if the current tab is an archive tab
		wxWindow* tab = notebook_archives->GetPage(notebook_archives->GetSelection());
		if (tab && tab->GetName() == "archive")
			((ArchivePanel*)tab)->handleAction(menu_id);	// Send action to current ArchivePanel
	}
}

/* ArchiveManagerPanel::updateBookmarkListItem
 * Updates the bookmark list item at <index>
 *******************************************************************/
void ArchiveManagerPanel::updateBookmarkListItem(int index) {
	// Only valid indices
	if (index < 0 || (unsigned)index >= theArchiveManager->numBookmarks())
		return;

	ArchiveEntry* entry = theArchiveManager->getBookmark(index);
	if (!entry)
		return;

	// Set item name
	list_bookmarks->setItemText(index, 0, entry->getName());
	list_bookmarks->setItemText(index, 1, entry->getParent()->getFilename());

	// Set item status colour
	if (entry->isLocked())
		list_bookmarks->setItemStatus(index, LV_STATUS_LOCKED);
	else switch (entry->getState()) {
		case 0:	list_bookmarks->setItemStatus(index, LV_STATUS_NORMAL); break;
		case 1:	list_bookmarks->setItemStatus(index, LV_STATUS_MODIFIED); break;
		case 2:	list_bookmarks->setItemStatus(index, LV_STATUS_NEW); break;
		default:list_bookmarks->setItemStatus(index, LV_STATUS_ERROR); break;
	}
}

/* ArchiveManagerPanel::refreshBookmarkList
 * Clears and rebuilds the bookmark list
 *******************************************************************/
void ArchiveManagerPanel::refreshBookmarkList() {
	// Clear the list
	list_bookmarks->ClearAll();

	// Add columns
	list_bookmarks->InsertColumn(0, "Entry");
	list_bookmarks->InsertColumn(1, "Archive");

	// Add each bookmark
	list_bookmarks->enableSizeUpdate(false);
	for (unsigned a = 0; a < theArchiveManager->numBookmarks(); a++) {
		list_bookmarks->addItem(a, wxEmptyString);
		updateBookmarkListItem(a);
	}

	// Update size
	list_bookmarks->enableSizeUpdate(true);
	list_bookmarks->updateSize();
}

/* ArchiveManagerPanel::deleteSelectedBookmarks
 * Deletes selected bookmarks from the list
 *******************************************************************/
void ArchiveManagerPanel::deleteSelectedBookmarks() {
	vector<int> selection = getSelectedBookmarks();

	// Don't continue if there are no selected items
	if (selection.size() == 0)
		return;

	// Clear selection
	list_bookmarks->clearSelection();

	// Remove bookmarks
	for (int a = selection.size()-1; a >= 0; a--) {
		theArchiveManager->deleteBookmark(a);
	}
}

/* ArchiveManagerPanel::goToBookmark
 * Open the bookmark in the entry panel
 *******************************************************************/
void ArchiveManagerPanel::goToBookmark(long index) {
	// Get the selected bookmark entry
	ArchiveEntry* bookmark = theArchiveManager->getBookmark(list_bookmarks->selectedItems()[0]);

	// Check it's valid
	if (!bookmark)
		return;

	// Open its parent archive in a tab
	openTab(bookmark->getParent());

	// Get the opened tab (should be an ArchivePanel unless something went wrong)
	wxWindow* tab = notebook_archives->GetPage(notebook_archives->GetSelection());

	// Check it's an archive panel
	if (!(s_cmp(tab->GetName(), "archive")))
		return;

	// Finally, open the entry
	((ArchivePanel*)tab)->openEntry(bookmark, true);
}


/*******************************************************************
 * ARCHIVEMANAGERPANEL EVENTS
 *******************************************************************/

/* ArchiveManagerPanel::onListArchivesChanged
 * Called when the user selects an archive in the open files list.
 * Updates the maps list with any maps found within the selected
 * archive
 *******************************************************************/
void ArchiveManagerPanel::onListArchivesChanged(wxListEvent& e) {
	// Get the selected archive
	Archive* selected_archive = theArchiveManager->getArchive(e.GetIndex());

	// Return if selection doesn't exist
	if (!selected_archive)
		return;

	populateMapList(selected_archive);
}

/* ArchiveManagerPanel::onListArchivesActivated
 * Called when the user activates an archive in the list.
 * Opens the archive in a new tab, if it isn't already open.
 *******************************************************************/
void ArchiveManagerPanel::onListArchivesActivated(wxListEvent& e) {
	// Open the archive tab, or create a new tab if it isn't already
	openTab(e.GetIndex());
}

/* ArchiveManagerPanel::onListMapsChanged
 * Called when the user selects a map in the maps list.
 *******************************************************************/
void ArchiveManagerPanel::onListMapsChanged(wxCommandEvent& e) {
}

/* ArchiveManagerPanel::onListMapsActivated
 * Called when the user activates a map in the maps list.
 * Opens the map in a new map editor window
 *******************************************************************/
void ArchiveManagerPanel::onListMapsActivated(wxListEvent& e) {
	//new MapEditorWindow();
	wxMessageBox("Map Editor is not implemented yet.");
}

/* ArchiveManagerPanel::onListArchivesRightClick
 * Called when the user right clicks an item on the archive list,
 * pops up a context menu
 *******************************************************************/
void ArchiveManagerPanel::onListArchivesRightClick(wxListEvent& e) {
	PopupMenu(menu_context_open);
}

/* ArchiveManagerPanel::onListRecentActivated
 * Called when the user activates an archive in the list.
 * Opens the archive in a new tab, if it isn't already open.
 *******************************************************************/
void ArchiveManagerPanel::onListRecentActivated(wxListEvent& e) {
	// Open the archive
	openFile(theArchiveManager->recentFile(e.GetIndex()));
	// Refresh the list
	refreshRecentFileList();
}

/* ArchiveManagerPanel::onListRecentRightClick
 * Called when the user right clicks an item on the archive list,
 * pops up a context menu
 *******************************************************************/
void ArchiveManagerPanel::onListRecentRightClick(wxListEvent& e) {
	PopupMenu(menu_context_recent);
}

/* ArchiveManagerPanel::onListBookmarksActivated
 * Called when the user activates an entry in the list.
 * Opens the entry and if needed its archive in a new tab, if it
 * isn't already open.
 *******************************************************************/
void ArchiveManagerPanel::onListBookmarksActivated(wxListEvent& e) {
	// Open the archive
	goToBookmark(e.GetIndex());
}

/* ArchiveManagerPanel::onListBookmarksRightClick
 * Called when the user right clicks an item on the bookmark list,
 * pops up a context menu
 *******************************************************************/
void ArchiveManagerPanel::onListBookmarksRightClick(wxListEvent& e) {
	PopupMenu(menu_context_bookmarks);
}

/* ArchiveManagerPanel::onMenu
 * Called when an item on the archive list context menu is selected
 *******************************************************************/
void ArchiveManagerPanel::onMenu(wxCommandEvent& e) {
	switch(e.GetId()) {

	// Open Archives menu
	case MENU_SAVE:		saveSelection();	break;	// Save
	case MENU_SAVEAS:	saveSelectionAs();	break;	// Save As
	case MENU_CLOSE:	closeSelection();	break;	// Close

	// Recent Files menu
	case MENU_OPEN:		openSelection();	break;	// Open
	case MENU_REMOVE:	removeSelection();	break;	// Remove

	// Bookmarks menu
	case MENU_GO:		goToBookmark();		break;	// Go To
	case MENU_DELETE:	deleteSelectedBookmarks();	break;	// Delete
	}
}

/* ArchiveManagerPanel::onArchiveTabChanged
 * Called when the user switches between archive tabs
 *******************************************************************/
void ArchiveManagerPanel::onArchiveTabChanged(wxAuiNotebookEvent& e) {
	// If an archive tab is selected, set the frame title accordingly
	int selection = notebook_archives->GetSelection();
	/*
	if (isArchivePanel(selection)) {
		Archive* archive = ((ArchivePanel*)notebook_archives->GetPage(selection))->getArchive();
		((wxFrame*)GetParent())->SetTitle(s_fmt("SLADE - %s", chr(archive->getFilename(false))));
	}
	else
		((wxFrame*)GetParent())->SetTitle("SLADE");
	*/

	if (isArchivePanel(selection)) {
		ArchivePanel* ap = (ArchivePanel*)notebook_archives->GetPage(selection);
		ap->currentArea()->updateStatus();
	}

	e.Skip();
}

/* ArchiveManagerPanel::onArchiveTabClose
 * Called when the user clicks the close button on an archive tab
 *******************************************************************/
void ArchiveManagerPanel::onArchiveTabClose(wxAuiNotebookEvent& e) {
	if (close_archive_with_tab) {
		tab_closing = true;
		Archive* archive = getArchive(e.GetInt());
		if (archive) closeArchive(archive);
		tab_closing = false;
	}
}

/* ArchiveManagerPanel::onAMTabChanged
 * Called when a different archive manager tab is selected
 *******************************************************************/
void ArchiveManagerPanel::onAMTabChanged(wxAuiNotebookEvent& e) {
	am_current_tab = notebook_tabs->GetSelection();
}
